/*-------------------------------------------------------------------------
 *  vmadump_threads.c:  Virtual Memory Area dump/restore routines
 *
 *  2004, The Regents of the University of California, through Lawrence
 *  Berkeley National Laboratory (subject to receipt of any required
 *  approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 *  Portions may be copyrighted by others, as may be noted in specific
 *  copyright notices within specific files.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: vmadump_threads.c,v 1.5 2006/07/11 23:06:13 phargrov Exp $
 *
 * THIS FILE SPECIFIC TO BLCR <http://ftg.lbl.gov/checkpoint>
 * It handles thread-management for dumps of multithreaded processes.
 * This file is shared by both the Linux 2.4 and 2.6 versions of vmadump.
 *-----------------------------------------------------------------------*/

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  #ifndef EXPORT_SYMTAB
    #define EXPORT_SYMTAB
  #endif
  #include <linux/config.h>
  #if defined(CONFIG_SMP) && ! defined(__SMP__)
    #define __SMP__
  #endif
  #if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
    #define MODVERSIONS
  #endif
  #if defined(MODVERSIONS)
    #include <linux/modversions.h>
  #endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>

#define __VMADUMP_INTERNAL__
#include "vmadump.h"
#include "blcr_config.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  #undef MOD_INC_USE_COUNT
  #define MOD_INC_USE_COUNT
  #undef MOD_DEC_USE_COUNT
  #define MOD_DEC_USE_COUNT
#endif 

#ifdef DEFINE_SPINLOCK
  DEFINE_SPINLOCK(thread_req_list_lock);
#else
  spinlock_t thread_req_list_lock = SPIN_LOCK_UNLOCKED;
#endif
LIST_HEAD(thread_req_list);

/* -------------------------------------------------------------------
 * Routines for tracking outstanding thread requests
 */
static struct vmadump_thread_req *
alloc_thread_req(struct task_struct *task, int threadop, struct file *file, int num_threads, int flags)
{
    struct vmadump_thread_req *req;
    
    req = kmalloc(sizeof(struct vmadump_thread_req), GFP_ATOMIC);
    if (req) {
	req->mm = task->mm;                    
        req->num_threads = num_threads;        
        atomic_set(&req->num_awake, num_threads);
        init_waitqueue_head(&req->thread_barrier);
        atomic_set(&req->num_frozen, num_threads);
        init_waitqueue_head(&req->exit_barrier);
        req->file = file;
	sema_init(&req->file_sem, 1);
        req->flags = flags;
        atomic_set(&req->ref_count, 1);
	INIT_LIST_HEAD(&req->req_list);
	req->type = threadop;
	req->error = 0;
    }

    return req;
}

/*
 * Note:  Must be called with thread_req_list_lock held
 */
static struct vmadump_thread_req *
find_thread_req(struct task_struct *task)
{
    struct list_head *tmp;
    struct vmadump_thread_req *req;

    list_for_each(tmp, &thread_req_list) {
        req = list_entry(tmp, struct vmadump_thread_req, req_list);
        if (req->mm == task->mm) {
	    atomic_inc(&req->ref_count);
	    return req;
	}
    }

    return NULL;
}

static struct vmadump_thread_req *
get_thread_req(int threadop, struct task_struct *task, struct file *file, 
               int num_threads, int flags)
{
    struct vmadump_thread_req *req;

    spin_lock(&thread_req_list_lock);
    req = find_thread_req(task);
    if (!req) {
	/* we're the first one's in.  Put a req in for everyone else */
	req = alloc_thread_req(task, threadop, file, num_threads, flags);
	if (!req) {
	    goto err;
	}

	/* ... put us on the list */
	list_add(&req->req_list, &thread_req_list);
    }

err:
    spin_unlock(&thread_req_list_lock);
    return req;
}


static void
put_thread_req(struct vmadump_thread_req *req)
{
    spin_lock(&thread_req_list_lock);
    if(atomic_dec_and_test(&req->ref_count)) {
        list_del(&req->req_list);
        kfree(req);
    }
    spin_unlock(&thread_req_list_lock);
}

long
vmadump_thaw_threads(struct file *file, struct pt_regs *regs, 
                     struct vmadump_threadundump_args *args)
{
    struct vmadump_thread_req *req;
    int pid;
    int retval;

    MOD_INC_USE_COUNT;

    req = get_thread_req(VMAD_THREAD_UNDUMP, current, file,
			 args->num_threads, args->flags);

    if (!req) {
	retval = -ENOMEM;
	goto err;
    }

    if (req->type != VMAD_THREAD_UNDUMP) {
	retval = -EAGAIN;
	goto err1;
    }

    if (atomic_read(&req->num_awake)) {
	if (atomic_dec_and_test(&req->num_awake)) {
	    retval = vmadump_thaw_proc(0, req->file, regs, 0, &pid);
	    req->error = retval;
	    memcpy(req->comm, current->comm, sizeof(req->comm));
	    wake_up(&req->thread_barrier);
	} else {
	    int num_awake;
	    /* ... wait to be woken up */
	    wait_event_interruptible(req->thread_barrier, !atomic_read(&req->num_awake));
	    /* ... test for the "good" wake condition */
	    if ((num_awake = atomic_read(&req->num_awake))) {
		retval = -EAGAIN;
		printk("vmadump[%d]: Warning: Thread awoke prematurely."
		       "%d threads still awake.\n", current->pid, num_awake);
	    } else {
		down(&req->file_sem);
		if (!req->error) {
		  retval = vmadump_thaw_proc(0, req->file, regs,
					     VMAD_DUMP_REGSONLY, &pid);
		  req->error = retval;
		  memcpy(current->comm, req->comm, sizeof(current->comm));
		} else {
		  retval = req->error;
		}
		up(&req->file_sem);
	    }
	}
    } else {
	printk("vmadump[%d]: Warning: Threadundump invoked after barrier\n",
	       current->pid);
	retval = -EAGAIN;
    }

    if (!retval) {
        args->old_pid = pid;
    }

err1:
    put_thread_req(req);
err:
    MOD_DEC_USE_COUNT;
    return retval;
}

long
vmadump_freeze_threads(struct file *file, struct pt_regs *regs,
                       struct vmadump_threaddump_args *args)
{
    struct vmadump_thread_req *req;
    int retval;

    MOD_INC_USE_COUNT;

    req = get_thread_req(VMAD_THREAD_DUMP, current, file,
			 args->num_threads, args->flags);

    if (!req) {
	retval = -ENOMEM;
	goto err;
    }

    if (req->type != VMAD_THREAD_DUMP) {
	retval = -EAGAIN;
	goto err1;
    }

    if (atomic_read(&req->num_awake)) {
	if (atomic_dec_and_test(&req->num_awake)) {
	    retval = vmadump_freeze_proc(0, file, regs, req->flags,
					 req->num_threads);
	    wake_up(&req->thread_barrier);
	} else {
	    /* ... wait to be woken up */
	    int num_awake;
	    /* XXX:  Ok, so there's no good reason to use a timeout
	     * there.  I'm leaving it as a just in case.  
	     * If this happens,
	     * there is serious potential for screwing things up in
	     * incredibly not nice ways, but at least we'll be able to 
	     * to unload vmadump for further debugging! */
	    wait_event_interruptible(req->thread_barrier, !atomic_read(&req->num_awake));
	    /* ... test for the "good" wake condition */
	    if ((num_awake = atomic_read(&req->num_awake))) {
		retval = -EAGAIN;
		printk("vmadump[%d]: Warning: Thread awoke prematurely."
		       "%d threads still awake.\n", current->pid, num_awake);
	    } else {
		down(&req->file_sem);
		retval = vmadump_freeze_proc(0, req->file, regs,
					     args->flags | VMAD_DUMP_REGSONLY,
					     req->num_threads);
		up(&req->file_sem);
	    }
	}
    } else {
	/* if num_threads is zero it means that either this req has
	 * happened already, or that it's in progress now.  In any case,
	 * this process no longer has any reason to be a part of this req.
	 */
	printk("vmadump[%d]: Warning: Threaddump invoked after barrier\n",
	       current->pid);
	retval = -EAGAIN;
    }

    /* now a second barrier, to make sure that we've all written */
    if (atomic_dec_and_test(&req->num_frozen)) {
        wake_up(&req->exit_barrier);
    } else {
        wait_event_interruptible(req->exit_barrier, !atomic_read(&req->num_frozen));
    }

err1:
    put_thread_req(req);
err:
    MOD_DEC_USE_COUNT;
    return retval;
}

EXPORT_SYMBOL(vmadump_freeze_threads);
EXPORT_SYMBOL(vmadump_thaw_threads);
